# coding: utf-8

"""
    assetic.APIhelper  (apihelper.py)
    Tools to assist with using Assetic API's and typical integration tasks
"""
from __future__ import absolute_import

import six
import webbrowser   #use for launching assetic
import threading    #use for launching assetic
import smtplib      #use for emailing

from ..configuration import Configuration
from ..api_client import ApiClient
from ..rest import ApiException

class APIHelper(object):
    """
    Class with generic tools to assist integrations
    """

    def __init__(self, api_client=None):
        config = Configuration()
        if api_client:
            self.api_client = api_client
        else:
            if not config.api_client:
                config.api_client = ApiClient()
            self.api_client = config.api_client
            
        self.logger = config.packagelogger

    def launch_url(self,url):
        """
        Launch a URL in a browser window
        Applications such as ArcMap and its python add-ins are
        sensitive to threading. To open a separate process it needs to be opened
        in it's own thread so it does not interfere with ArcMap.
        The code below will open the url in a new tab in a separate thread.
        :param url: The URL to launch
        """
        #open the url in a new tab in a separate thread.
        threading.Thread(target=webbrowser.open, args=(url,0)).start()      
    
    def launch_assetic_dimension(self,assetguid,componentguid):
        """
        Open assetic and display the dimension page for the component
        :param assetguid: asset guid
        :param componentguid: component guid
        """
        url = "{0}/Assets/{1}/Component/Component/{2}/CPDimension/Default/"\
              .format(self.api_client.host,assetguid,componentguid)
        self.launch_url(url)
        
    def launch_assetic_workorder(self,wkoid):
        """
        Open assetic and display the work order
        :param wkoid: work order guid or friendly ID
        """
        url = "{0}/Maintenance/Operational/Plan/EditWO/?Id={1}".format(
            self.api_client.host,wkoid)
        self.launch_url(url)

    def launch_assetic_asset(self,assetid):
        """
        Open assetic and display the asset
        :param assetid: asset guid or friendly ID
        """
        url = "{0}/Assets/ComplexAsset/{1}".format(
            self.api_client.host,assetid)
        self.launch_url(url)
        
    def launch_assetic_work_request(self,wrid):
        """
        Open assetic and display the asset
        :param wrid: work request guid or friendly ID
        """
        url = "{0}/Maintenance/Operational/Request/EditWR/?Id={1}".format(
            self.api_client.host,wrid)
        self.launch_url(url)

    def smtp_mailer(self,mailhost, port, username, password, fromaddr, toaddrs, msg,isthread = False):  
        """
        Send an email via SMTP using TLS (port 587) or SSL (port 465) or
        no auth (port 25)
        Better to use TLS if possible since TLS supercedes SSL
        :param mailhost: the SMTP host
        :param port: SMTP port, expect one of 587 (TLS), 465 (SSL), or 25
        :param username: username.  May be None
        :param password: password. May be None
        :param fromaddr: 'From' email address
        :param toaddrs: 'To' email addresses, comma separated
        :param msg: email subject and body (body separated by 2 carriage returns
        from subject
        :param ishandler: True if this is an SMTP logging handle.
        Need to know so then this method won't write exceptions sending th
        email becuase the hanlder picks up the error and it causes an error loop
        """

        smtp = None
        if port == "587":
            try:
                smtp = smtplib.SMTP(mailhost, port)
            except Exception as ex:
                if isthread == False:
                    msg = "Unable to initialise TLS smtp variable. "\
                          "Host: {0}; Port {1}; Error {2}".format(
                              mailhost, port,str(ex))
                    self.logger.error(msg)
                return 1
        elif port == "465":
            try:
                smtp = smtplib.SMTP_SSL(mailhost, port)
            except Exception as ex:
                if isthread == False:
                    msg = "Unable to initialise SSL smtp variable. "\
                          "Host: {0}; Port {1}; Error {2}".format(
                              mailhost, port,str(ex))
                    self.logger.error(msg)
                return 1
        elif port == "25":
            #try simple email module
            try:
                smtp = smtplib.SMTP(mailhost, port)
            except Exception as ex:
                if isthread == False:
                    msg = "Unable to initialise smtp variable. "\
                      "Host: {0}; Port {1}; Error {2}".format(
                          mailhost, port,str(ex))
                    self.logger.error(msg)
                return 1
        if smtp == None:
            if isthread == False:
                self.logger.error("email protocol not configured/supported. "\
                          "Host: {0}; Port {1}".format(mailhost, port))
            return 1
        try:
            if username:
                smtp.ehlo() #possibly optional for SSL.  Need for TLS
                if port =="587":
                    # for tls need these extra lines
                    smtp.starttls()
                    smtp.ehlo()# for tls add this line
                smtp.login(username, password)
            #Now send the actual email
            smtp.sendmail(fromaddr, toaddrs, msg)
            smtp.quit()
        except Exception as ex:
            if isthread == False:
                self.logger.error("Error sending email. "\
                          "Host: {0}; Port {1}; Error: {2}".format(
                              mailhost, port,str(ex)))
            return 1            
        return 0
 

