# coding: utf-8

"""
    Assetic.bulk_export.py
    Tools to facilitate bulk export of data from assetic and
    syncing exported data to a local sql server database
    
    Checks if there are files to download from from Export All task
    Downloads the file and saves locally to then process
    Uses file system to record pending & processed export jobs
    Data is written to a temp db table and then the 'merge' SQL is used
    to update/insert records in the permanant table
    Will create the tables as required (TODO check columns exist)
    Writes to a log table the changed record (TODO expand the output)
"""
#from __future__ import absolute_import

import pyodbc
import sys
import datetime
import dateutil.parser
import string  #(use string.lower instead of str.lower for py2.7)

import os
import shutil
from contextlib import contextmanager
from pprint import pformat
import six
import time
import codecs
import csv

#    import cStringIO

from ..configuration import Configuration
from ..api_client import ApiClient
from ..rest import ApiException
from assetic.apis import DocumentApi
from assetic.apis import SearchApi
from assetic.apis import BackgroundWorkerApi
from assetic.tools import BulkProcesses

class SyncToLocalProcesses(object):
    """
    Class to manage processes that sync Assetic search data to a local DB
    """

    def __init__(self, api_client=None):
        config = Configuration()
        if api_client:
            self.api_client = api_client
        else:
            if not config.api_client:
                config.api_client = ApiClient()
            self.api_client = config.api_client
            
        self.logger = config.packagelogger


    @property
    def workingdir(self):
        """
        Gets user agent.
        """
        if self.workingdir == None:
            workingdir = os.getcwd()
        return self.workingdir

    @workingdir.setter
    def workingdir(self, value):
        """
        Sets user agent.
        """
        if os.path.isdir(value) == True:
            self.workingdir = value
        else:
            raise ApiException(
                status=0,
                reason="Invalid workingdir set: `{0}`"
                .format(value)
            )
        
    def create_export_task(self,profile,dbtools,incremental):
        """
        Given a search profile initiate a bulk export job
        Param1: The search profile guid
        Returned: return code 0=success,1=error
        """

        ##have added a new field to assetic_sync_manager Feb2017
        ##check to see it is there and create if not
        returncode,existingcolumns = dbtools.get_columns_for_table("assetic_sync_manager")
        if returncode !=0:
            return returncode
        if "replacespaces" not in existingcolumns:
            ##need to create
            sql = "alter table assetic_sync_manager add replacespaces bit"
            returncode = dbtools.execute_single(sql)
        if returncode !=0:
            self.logger.error('Table assetic_sync_manger missing field "replacespaces"')
            return returncode
        ##Add spacedelimiter field to allow an underscore or other delimiter
        ##to replace the space.  Must also set replacespaces = True to use
        if "spacedelimiter" not in existingcolumns:
            ##need to create
            sql = "alter table assetic_sync_manager add "\
                  "spacedelimiter [nvarchar](50)"
            returncode = dbtools.execute_single(sql)
        if returncode !=0:
            self.logger.error('Table assetic_sync_manger missing field '\
                              '"spacedelimiter"')
            return returncode        
        #first see how may rows are returned, if less than 10,000
        #can use the search get rather than batch - quicker as no
        #background worker process
        ##write task details to database
        sql = """insert into assetic_sync_manager
            (oid,taskid, profileid,profilename,tablename,keyfield,logtable,
            useinternalnames,replacespaces,spacedelimiter,status,
            date_initiated,date_last_modified)
            values (NEWID(),?,?,?,?,?,?,?,?,?,?,getdate(),getdate())"""

        #TODO record current datetime as reference

        searchapi = SearchApi(self.api_client)

        totalrecords=0
        ##if incremental is false, use full export
        if incremental == False:
            kw = {'request_params_id':profile.profileguid,
                'request_params_page':1,
                #'request_params_filters':'',
                'request_params_page_size':1}

            ##Get first page of results
            try:
                sg = searchapi.search_get(**kw)
            except ApiException as e:
                self.logger.error('Status {0}, Reason: {1}'.format(
                    e.status,e.reason))
                returncode = int(e.status)
                return returncode 
            totalrecords=sg.get('TotalResults')

        ##Temporary hardcode to ensure always export all - until have
        ##capacity to reliably search by last modified date
        totalrecords=0
        
        if 0 < totalrecords < 9900: 
            #use standard search api to get data
            csvstring = run_search_for_pages(profile.profileguid)
            if csvstring == None:
                return 0
            metadata = get_search_columns(profile.profileguid)
            returncode = csv_string_to_db(self,csvstring,profile,dbtools,metadata)
            if returncode != 0:
                return returncode            
            ##update as processed
            record = "not required", profile.profileguid,profile.profilename,\
                     profile.tablename,profile.keyfield,profile.logtable,\
                     profile.useinternalnames,profile.replacespaces,\
                     profile.spacedelimiter,"Complete"
            returncode = dbtools.execute_single(sql,record)
        else:
            ##else create export task
            try:
                taskid = searchapi.search_post_export_profile(
                    profile.profileguid)
            except ApiException as e:
                self.logger.error('Status {0}, Reason: {1}'.format(
                    e.status,e.reason))
                returncode = int(e.status)
                return returncode
            
            if taskid != None:
                record = taskid, profile.profileguid,profile.profilename,\
                         profile.tablename,profile.keyfield,profile.logtable,\
                         profile.useinternalnames,profile.replacespaces,\
                         profile.spacedelimiter,"In Progress"
                returncode = dbtools.execute_single(sql,record)
            else:
                returncode = 1
                
        return returncode
    

    def get_export_docid(self,exporttaskguid):
        """
        Given a export job GUID check status
        Param1: the guid of the export task
        Returned: A tuple (2 fields)
        1. The status or None if profile not found
        2. docid if ready for download, otherwise None
        """
        bwapi = BackgroundWorkerApi(self.api_client)
        try:
            response = bwapi.background_worker_get(exporttaskguid)
        except ApiException as e:
            msg="Create background worker task for {0}".format(exporttaskguid)
            self.logger.error('{0}: Status {0}, Reason: {1}'.format(
                msg,e.status,e.reason))
            return None,None            
        status = response.get("Status")
        docid = None

        if status == "Completed":
            docid = response.get("DocumentId")
        return status,docid

    def get_document_and_save(self,docid,savefolder):
        """
        Downloads the export file to disk    
        Returned: the csv file path(full path) for a given document id
        """
        
        ##download the file
        bulkprocesses = BulkProcesses(self.configfile)
        csvfile = bulkprocesses.document_get(savefolder,docid)
        ##return the file name
        return csvfile

    def build_assetic_sync_manager_table(self,dbserver,dbname):
        """
        Builds the db table used to manager the sync process
        Assumes trusted connection to database
        param dbserver: the name of the SQL Server instance
        param dbname: the name of the the SQL Server database
        Returned: Success (Boolean) - True if no error 
        """
        dbtools = DB_Tools(dbserver,dbname)
        sqlchk = "select object_id('assetic_sync_manager') as objectid"
        sqlcreate = """CREATE TABLE assetic_sync_manager(
            [oid] [uniqueidentifier],            
            [taskid] [nvarchar](50),
            [profileid] [nvarchar](50),
            [profilename] [nvarchar](100),
            [tablename] [nvarchar](50),
            [keyfield] [nvarchar](50),
            [logtable] [nvarchar](50),
            [documentid] [nvarchar](50),
            [useinternalnames] bit,
            [replacespaces] bit,
            [spacedelimiter] [nvarchar](50),
            [date_initiated] datetime,
            [date_last_modified] datetime,
            [status] [nvarchar](50))"""
        returncode, result = dbtools.execute_select(sqlchk)
        if returncode == 0 and result[0][0] == None: 
            returncode = dbtools.execute_single(sqlcreate)
        return returncode

    def build_table_from_search(self,dbtools,metadata,profile,printonly=False):
        """
        Builds the sync target db table based on search fields
        Optionally print creation SQL instead of creating
        Can use standalone.  Is also used when sync is peformed
        Param dbtools: instance of dbtools
        Param metadata: instance of search profile metadata.
        if None then profile ID will be used to get the metadata from api
        Assetic3IntegrationRepresentationsSearchMetadataRepresentationList
        Param profile: insance of search profile definition
        Param printonly: If true the column name is based on the
        user visible field label in export, else internal field name is used
        default is False
        If true then table is not created but SQL for table create is.
        Allows manual edit of datatypes and manual creation of table.
        Returned: 0=Success, 1=error
        """
        if metadata == None:
            metadata = self.get_search_columns(profile.profileguid)
        if metadata == None:
            return 1
        sqlchk = "select object_id('{0}') as objectid".format(profile.tablename)
        resourcelist = metadata.get("ResourceList")
        columnddl = []
        for columndef in resourcelist:
            #by default set column name to user friendly column header
            #in export file
            colname = "[{0}]".format(columndef.get("Label"))
            if profile.useinternalnames == True:
                #change column name to assetic internal field name for the label
                colname = "[{0}]".format(columndef.get("Name"))
            elif profile.replacespaces == True:
                ##replace spaces in field name with delimiter
                if profile.spacedelimiter == None:
                    ##remove whitespaces from label
                    profile.spacedelimiter = ""
                #replace whitespaces from Label names
                colname = colname.replace(" ",profile.spacedelimiter)
            #set datatype
            datatype = "nvarchar(500)"    #default datatype - e.g System.String
            typedef = columndef.get("Type").lower()
            if typedef.find('integer') > 0:
                datatype = 'int'
            elif typedef.find('double') > 0:
                datatype = 'float'
            elif typedef.find('decimal') > 0:
                datatype = 'float'
            elif typedef.find('datetime') > 0:
                datatype = 'datetime'
            elif typedef.find('bool') > 0:
                datatype = 'bit'
            #set NULL/NOT NULL
            allownull = "NULL"
            if colname.lower() == profile.keyfield.lower():
                allownull = "NOT NULL"
            
            columnddl.append("{0} {1} {2}".format(colname,datatype,allownull))

        #finalise create string                 
        sqlcreate = "create table {0}({1})".format(profile.tablename,
                        ",".join(columnddl))
        #check to see if table exists
        returncode, result = dbtools.execute_select(sqlchk)
        
        if returncode == 0 and result[0][0] == None:
            if printonly == True:
                print(sqlcreate)
            else:
                #create table
                returncode = dbtools.execute_single(sqlcreate)
        elif returncode == 0 and result[0][0] != None:
            self.logger.debug("Target table [{0}] already exists".format(profile.tablename))
        return returncode
    
    def get_search_columns(self,profileid):
        """
        Use the api to get column mappings for search profile
        Param profileid: search profile guid
        Returned: array of SearchProfileColumnRepresentation or None if error or
        profile not found
        """
        #client = assetic.ApiClient(self.conf.host,"Authorization",self.auth)
        searchapi = SearchApi(self.api_client)
        try:
            response = searchapi.search_get_profile_metadata(profileid)
        except ApiException as e:
            self.logger.error('Status {0}, Reason: {1}'.format(
                e.status,e.reason))
            returncode = int(e.status)
            return None
        return response
               
    def get_document(self,docid):
        """
        Gets document from Assetic, but doesn't write to disk
        Param docid: document ID to retrieve
        Returned: the data component of the document only. File not saved
        """
        #client = assetic.ApiClient(self.conf.host,"Authorization",self.auth)
        docapi = DocumentApi(self.api_client)
        try:
            data = docapi.document_get_document_file(docid)
        except ApiException as e:
            self.logger.error('Status {0}, Reason: {1}'.format(
                e.status,e.reason))
            returncode = int(e.status)
            return None        
        ##ensure encoding is ok
        datadec = self.decode_bom(data)
        return datadec
    
    def decode_bom(self,data):
        """
        Remove BOM from data
        Param data: Data to decode
        Returned: input data decoded to UTF-8
        """
        if sys.version_info < (3,0):
##            for encoding in "utf-8-sig", "utf-16":
##                try:
##                    self.logger.debug("changing codec from BOM_UTF8 to UTF_8")
##                    return data.decode(encoding)
##                except UnicodeDecodeError:
##                    continue
##            return data.decode("latin-1") # will always work
            if data.startswith(codecs.BOM_UTF8):
                self.logger.debug("removing BOM_UTF8 codec from document start")
                data = data.strip(codecs.BOM_UTF8)
            return data            
        else:
            #v3.  Remove BOM if present
            if data.startswith(codecs.BOM_UTF8.decode("utf8")):
                self.logger.debug("removing BOM_UTF8 codec from document start")
                data = data.strip(codecs.BOM_UTF8.decode("utf8"))
            return data
    
    def get_active_task_doclist_db(self,dbtools):
        """
        get active tasks frm assetic_sync_manager table using filter
        status='inprogress'
        Param dbtools: dbtools object        
        Returned: a list of active tasks and document id
        """

        ##have added a new field to assetic_sync_manager Feb2017
        ##check to see it is there and create if not
        returncode,existingcolumns = dbtools.get_columns_for_table("assetic_sync_manager")
        if returncode !=0:
            return returncode
        if "replacespaces" not in existingcolumns:
            ##need to create
            sql = "alter table assetic_sync_manger add replacespaces bit"
            returncode = dbtools.execute_single(sql)
        if returncode !=0:
            self.logger.error('Table assetic_sync_manger missing field "replacespaces"')
            return returncode
        ##Add spacedelimiter field to allow an underscore or other delimiter
        ##to replace the space.  Must also set replacespaces = True to use
        if "spacedelimiter" not in existingcolumns:
            ##need to create
            sql = "alter table assetic_sync_manager add "\
                  "spacedelimiter [nvarchar](50)"
            returncode = dbtools.execute_single(sql)
        if returncode !=0:
            self.logger.error('Table assetic_sync_manger missing field '\
                              '"spacedelimiter"')
            return returncode        
        activetasks = []

        #get all tasks
        sql = """select convert(nvarchar(36),oid),taskid, profileid,profilename,
            tablename,keyfield,logtable,useinternalnames,replacespaces,
            spacedelimiter from assetic_sync_manager where status='In Progress'
            order by [date_initiated] asc"""
        returncode,tasks = dbtools.execute_select(sql)
        if returncode != 0:
            return None
        
        ##format to task representation
        for task in tasks:
            ##first check if export document ready
            taskstatus,docid = self.get_export_docid(task[1])
            if docid != None:
                tr = ExportTaskRepresentation()
                pr = SearchProfileRepresentation()
                tr.oid = task[0]
                tr.taskguid = task[1]
                tr.documentid = docid
                pr.profileguid = task[2]
                pr.profilename = task[3]
                pr.tablename = task[4]
                pr.keyfield = task[5]
                pr.logtable = task[6]
                pr.useinternalnames = task[7]
                pr.replacespaces = task[8]
                pr.spacedelimiter = task[9]
                tr.profile = pr
                activetasks.append(tr)
            else:
                msg = "Export file for profile [{0}] is not ready for download"\
                      ".  The status of the background worker process is [{1}]"\
                      ".  Please retry later when the file is ready".format(
                          task[3],taskstatus)
                self.logger.info(msg)
        return activetasks

    def csvfile_to_db(self,csvfile,profile,dbtools):
        """
        For a given csv file sync data to db.  Assumes a header row
        Returned: 0 if success, else 1
        """

        dbtable = profile.tablename
        keyfield = profile.keyfield
        logtable = profile.logtable
        
        #get list of columns
        bIsHeader = True
        allrows = []
        saved_row = []

        with open(csvfile, 'rt', encoding='utf-8', newline = '') as csvfile:
            readCSV = csv.reader(csvfile,delimiter=',')
            for row in readCSV:
                if bIsHeader == True:
                    bIsHeader = False
                    columns = row
                else:
                    #cater for carriage returns in a column
                    if len(saved_row):
                        row_tail = saved_row.pop()  #gets last item in the list
                        row[0] = row_tail + '\r' +row[0]  # reconstitute field broken by newline
                        row = saved_row + row       # and reassemble the row (possibly only partially)
                    if len(row) >= len(columns):
                        allrows.append(row)
                        saved_row = []
                    else:
                        saved_row = row
        returncode = dbtools.data_sync_to_db(allrows,columns,dbtable,keyfield,logtable)
        return returncode
        
    def csv_string_to_db(self,csvstring,profile,dbtools,metadata):
        """
        For a given single column array sync data to db.
        Assums data in column is comma delimited, but no need to split
        Assumes a header row
        param csvstring: a string of csv data (multiple lines)
        param profile: instance of SearchProfileRepresentation
        param dbtools: instance of dbtools
        param metadata: instance of search profile column metadata
        Returned: 0 if success, else 1
        """

        dbtable = profile.tablename
        keyfield = profile.keyfield
        logtable = profile.logtable
                    
        #get list of columns
        bIsHeader = True
        allrows = []
        saved_row = []
        csvarray = csvstring.splitlines()
##        if sys.version_info < (3,0):
##            #reader = UnicodeReader(csvarray)
##            reader = csv.reader(csvarray)
##        else:
##            reader = csv.reader(csvarray)
        reader = csv.reader(csvarray)
        for row in reader:
            #row.append(row.pop().replace('\n', ' ').replace('\r', ''))
            #if rowcnt <= 100:
            #    #only show first 100 rows
            #    rowcnt = rowcnt + 1
            #    self.logger.debug("CSV row {2}/100: {0}, count {1}".format(row,len(row),rowcnt))
            if bIsHeader == True:
                bIsHeader = False
                if profile.useinternalnames == True:
                    columnmap = metadata.get("ResourceList")
                    columns=[]
                    #switch the column headers with internal names
                    for header in row:
                        #find header in metadata and get internal name
                        for columnset in columnmap:
                            if columnset.get('Label') == header:
                                columns.append(columnset.get('Name'))
                                break
                    self.logger.debug("Using internal field names for column names")
                elif profile.replacespaces == True:
                    ##replace spaces in field name with delimiter
                    if profile.spacedelimiter == None:
                        ##remove whitespaces from label
                        profile.spacedelimiter = ""
                    columns = []
                    for header in row:
                        columns.append(header.replace(" ",profile.spacedelimiter))
                    self.logger.debug("Using label names without whitespaces from\
                        column names")
                else:
                    columns = row
                #self.logger.debug("CSV file columns: {0}, count {1}".format(columns,len(columns)))
            else:
                #cater for carriage returns in a column
                if len(saved_row):
                    row_tail = saved_row.pop()  #gets last item in the list
                    row[0] = row_tail + '\r' +row[0]  # reconstitute field broken by newline
                    row = saved_row + row       # and reassemble the row (possibly only partially)
                if len(row) >= len(columns):
                    decodedrow = row
                    if sys.version_info < (3,0):
                        #need to encode as unicode for python 2 otherwise issues
                        #with special characters
                        try:
                            decodedrow = [field.decode("utf8") for field in row]
                        except:
                            decodedrow = row
                    allrows.append(decodedrow)
                    saved_row = []
                else:
                    saved_row = row
        returncode = dbtools.data_sync_to_db(allrows,columns,dbtable,keyfield,logtable)
        return returncode
    def run_search_for_pages(searchguid):
        """
        Where a search < 10,000 records use the search api rather than
        export api.
        Assums data in column is comma delimited, but no need to split
        Assumes a header row
        Param searchguid: the search profile guid
        Param startpage: the first page to search from
        Param pagelimit: the page to search up to
        Returned: search data as csv or None
        """
        ##create instance of search api
        sapi = SearchApi(self.client)

        ##set parameters for search
        numpagesize = 500  #max number of pagess
        startpage = 1
        pagelimit = 20
        kw = {'request_params_id':searchguid,
            'request_params_page':startpage,
            'request_params_page_size':numpagesize}

        ##Get first page of results
        try:
            sg = sapi.search_get(**kw)
        except ApiException as e:
            msg = "Run search for pages"
            self.logger.error('{0}: Status {1}, Reason: {2}'.format(
                msg,e.status,e.reason))
            returncode = int(e.status)
            return None
        morepages = True
        totalresults=int(sg.get('TotalResults'))
        if totalresults == 0:
            morepages = False
            return None
        numpages = int(sg.get('TotalPages')) - (startpage-1)
        self.logger.debug('Total Pages: {0}'.format(numpages))
        if pagelimit > int(numpages):
            ##page limit defines the number of pages we'll process before adding
            ##to DB
            pagelimit = numpages
        if numpages <= pagelimit:
            morepages = False
        endpage = (startpage-1) + pagelimit
        self.logger.debug("Start page: {0}, End page: {1}, pages left: {2}".format(startpage, endpage,numpages))
        
        ##get actual data from nested output
        resourcelist = sg.get('ResourceList')
        resource = resourcelist[0]
        data = resource.get('Data')
        ##Get a list of columns
        mappings = resource.get('Mappings')
        columns = list(mappings.keys())
        self.logger.debug("first page, columns are: {0}".format(columns))
        ##Copy as alldata becuase we will be appending to this
        alldata = data

        ##Write list of columns to the 'all' list as this may grow
        allcolumns = columns

        ##Now loop through remaining pages
        if numpages > 1:
            for pagenum in range(startpage + 1,endpage +1):
                ##set page number to get
                kw['request_params_page']=pagenum
                self.logger.debug('Page: {0}'.format(kw['request_params_page']))

                ##Now get results for this page
                try:
                    sg = sapi.search_get(**kw)
                except ApiException as e:
                    msg = "Run search for pages - additional pages"
                    self.logger.error('{0}: Status {1}, Reason: {2}'.format(
                        msg,e.status,e.reason))
                    returncode = int(e.status)
                    return None
                
                totalresults=int(sg.get('TotalResults'))
                if totalresults == 0:
                    morepages = False
                    break
                ##get actual data from nested output
                resourcelist = sg.get('ResourceList')
                resource = resourcelist[0]
                data = resource.get('Data')
                mappings = resource.get('Mappings')
                columns = list(mappings.keys())

                #Get column list for this page - there may be new columns
                self.logger.debug("in page loops, columns are: {0}".format(columns))
                if sys.version_info < (3,0):
                    ##Add new column list to 'allcolumns', will get unique list later
                    allcolumns = allcolumns + columns
                    #append new data to 'alldata'
                    alldata = alldata + data
                else:
                    ##merge column list sets
                    allcolumns.extend(columns)
                    #append new data to 'alldata'
                    alldata.extend(data)

                if startpage + pagenum > 20:
                    ##catchall escape
                    morepages = False
                    break
        csvrow = []
        for i_r in alldata:
        #map data to column list by key to avoid issues with column order
            #allrows.append(tuple(map( lambda x: i_r.get( x, "" ), columns )))
            csvrow.append = map( lambda x: i_r.get( x, "" ), columns ).join(",")
        alldata = columns.join(",") + "\n" + csvrow.join("\n")
           
        return alldata


    def initiate_export_process(self,searchid,dbtools):
        """
        For a given search profile initiate the bulk export
        Returned: True = Success, False = Error
        Also create a file named after the task id.  This file is in the
        'inprogress' folder.
        Later we can check this folder for 'inprogress' tasks
        OBSOLETE
        """
        taskid = self.create_export_task(searchid,dbtools)
        if taskid == None:
            return False
        ##create an empty file in the working 'in progress' folder
        ##filename is the taskid
        #file = self.workingdir + "/inprogress/" + taskid 
        #try:
        #    open(file,'w').close()
        #except:
        #    self.logger.error("Unable to save file to {0}".format(file))
        #    return False
        return True

    def get_csv_if_ready_process(self):
        """
        Look in the working folder 'inprogress' for file
        The file name is the guid of a job.  Check if job is complete
        and if so, download the file to the 'awaitingsync' folder
        OBSOLETE
        """
        sourcedir = self.workingdir + "/inprogress"
        downloaddir = self.workingdir + "/csvexports"
        processedjobs = self.workingdir + "/processedjobs"
        files = os.listdir(sourcedir)
        for filename in files:
            csvfile = get_document_and_save(filename,downloaddir)
            if csvfile != None:
                shutil.move(sourcedir + "/" + filename,processedjobs + "/" + filename)
            else:
                shutil.move(sourcedir + "/" + filename,downloaderror + "/" + filename)


    def dbsync_initiate(self,dbserver,dbname,profiles,incremental=True):
        """
        For each profile in profiles array create new exports and log export
        task ID in db for later 'finalisation' step
        With this method there are no files used to manage process
        By default it is intended that it will only gets records modified
        since last sync date, however this is not yet implemented. All records
        are therefore downloaded.  The parameter incremental will at a future
        point be utilised to control this behaviour
        Param dbserver: database instance (SQL Server)
        Param dbname: database name.  Assumes trusted connection
        Param profiles: an array of SearchProfileRespresentation instances
        Param incremental: True/False.  Apply incremeal export if True (default).  
        Default=True, however not implemented yet.
        Returns: no errors = 0, error=>0. 
        """
        self.logger.info("Starting dbsync_initiate")
        dbtools = DB_Tools(dbserver,dbname)
        ##loop through the list of profiles and create exports
        errorcode = 0
        for profile in profiles:
            errorcode = self.create_export_task(profile,dbtools,incremental)
            if errorcode != 0:
                self.logger.info("Exiting dbsync_finalise with error")
                return errorcode
        self.logger.info("Finished dbsync_initiate")
        return errorcode
                
    def dbsync_finalise(self,dbserver,dbname):
        """
        Finalise sync by checking for documents created by background worker
        process and downloading those files to an array (not to disk)
        The downloaded data is synced and the task flagged as complete
        Param1: database instance (SQL Server)
        Param2: database name.  Assumes trusted connection
        Returns: no errors = 0, error=>0. 
        """
        self.logger.info("Starting dbsync_finalise")
        returncode = 0
        dbtools = DB_Tools(dbserver,dbname)
        ##get list of in-progress tasks from db
        tasks = self.get_active_task_doclist_db(dbtools)
        if tasks == None:
            self.logger.info("No active task exports ready for download")
        else:
            ##Prepare SQL for updating manager table on a per profile basis
            sql = """update assetic_sync_manager set documentid = ?,
                    status = ?, date_last_modified=getdate()
                    where oid=CONVERT(uniqueidentifier,?)"""
            ##get generated document ID for task
            for task in tasks:
                ##get document data
                docid = task.documentid
                docdata = self.get_document(docid)
                if docdata == None:
                    msg = "[{0}] export file contains no data."\
                          "Please check search profile.  Task ID:{1}".format(
                        task.profilename)
                    self.logger.warning(msg)
                else:
                    profile = task.profile
                    self.logger.info("Processing document {0} for profile {1} "
                        "for task {2}".format(docid,profile.profilename,
                        task.taskguid))
                    ##get column mappings and data types for search profile
                    datadef = self.get_search_columns(profile.profileguid)
                    ##build target table (method first checks if existing)
                    returncode = self.build_table_from_search(dbtools,
                                datadef,profile,False)
                    if returncode != 0:
                        return returncode
                    returncode = self.csv_string_to_db(docdata,profile,dbtools,
                                datadef)

                    ##update assetic_sync_manager
                    if returncode == 0:
                        status = "Processed"
                    else:
                        status = "Error"
                    vals = [docid,status,task.oid]
                    returncode = dbtools.execute_single(sql,vals)
                    if returncode != 0:
                        self.logger.error("Error updating"
                                    " assetic_sync_manager_table")
        self.logger.info("Finished dbsync_finalise")
        return returncode

    def dbsync_csvfile_process(self,dbserver,dbname,profiles):
        """
        Look in the working folder 'csvexports' for file
        The file name is the profile name of the job.
        Upload to database and move to csvsynced folder
        """
        sourcedir = self.workingdir + "/inprogress/"
        downloaddir = self.workingdir + "/csvexports"
        processedcsv = self.workingdir + "/processedcsv"
        errorcsv = self.workingdir + "/errorcsv"


        files = os.listdir(downloaddir)
        if len(files) == 0:
            return
        dbtools = DB_Tools(dbserver,dbname)
        for filename in files:
            self.logger.debug(filename)
            fullfilename = downloaddir + "/" + filename
            for profile in profiles:
                if profile.profilename + ".csv" == filename:
                    self.logger.debug("profile: {0},file: {1}".format(profile.profilename,filename))
                    success = self.csvfile_to_db(fullfilename,profile,dbtools)
                    #dt = datetime.datetime.utcnow().isoformat()
                    dt = str(int(time.time()))
                    filename = dt + "_" + filename
                    if success:
                        shutil.move(fullfilename,processedcsv + "/" + filename)
                    else:
                        shutil.move(fullfilename,errorcsv + "/" + filename)

class SearchProfileRepresentation(object):
    """"
    A structure for defining table metadata and relationships between
    search profile names, id's and tables
    """
    def __init__(self,profileguid=None,profilename=None,tablename=None,keyfield=None,logtable=None,useinternalnames=False,replacespaces=False,spacedelimiter=None):
        """
        SearchProfileRepresentation - a model defining metadata for syncing a
        search profile to a local database table via bulk export api
        """
        self.fieldtypes = {
            "profileguid": "str",
            "profilename": "str",
            "tablename": "str",
            "keyfield": "str",
            "logtable": "str",
            "useinternalnames":"bool",
            "replacespaces":"bool"
        }
        self._profileguid = profileguid
        self._profilename = profilename
        self._tablename = tablename
        self._keyfield = keyfield
        self._logtable = logtable
        self._useinternalnames = useinternalnames
        self._replacespaces = replacespaces
        self._spacedelimiter = spacedelimiter
        
    @property
    def profileguid(self):
        return self._profileguid
    @profileguid.setter
    def profileguid(self,profileguid):
        self._profileguid = profileguid
    
    @property
    def profilename(self):
        return self._profilename
    @profilename.setter
    def profilename(self,profilename):
        self._profilename = profilename

    @property
    def tablename(self):
        return self._tablename
    @tablename.setter
    def tablename(self,tablename):
        self._tablename = tablename

    @property
    def keyfield(self):
        return self._keyfield
    @keyfield.setter
    def keyfield(self,keyfield):
        self._keyfield = keyfield

    @property
    def logtable(self):
        return self._logtable
    @logtable.setter
    def logtable(self,logtable):
        self._logtable = logtable

    @property
    def useinternalnames(self):
        return self._useinternalnames
    @useinternalnames.setter
    def useinternalnames(self,useinternalnames):
        self._useinternalnames = useinternalnames

    @property
    def replacespaces(self):
        return self._replacespaces
    @replacespaces.setter
    def replacespaces(self,replacespaces):
        self._replacespaces = replacespaces

    @property
    def spacedelimiter(self):
        return self._spacedelimiter
    @spacedelimiter.setter
    def spacedelimiter(self,spacedelimiter):
        self._spacedelimiter = spacedelimiter
                              
    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in six.iteritems(self.fieldtypes):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

class ExportTaskRepresentation(object):
    """"
    A structure for defining table metadata and relationships between
    search profile names, id's and tables
    """
    def __init__(self,oid=None,taskguid=None,documentid=None,profile=None):
        """
        SearchProfileRepresentation - a model defining metadata for syncing a
        search profile to a local database table via bulk export api
        """
        self.fieldtypes = {
            "oid": "str",
            "taskguid": "str",
            "documentid": "str",
            "profile": "SearchProfileRepresentation"
        }
        self._oid = oid
        self._taskguid = taskguid
        self._documentid = documentid
        self._profile = profile

    @property
    def oid(self):
        return self._oid
    @oid.setter
    def oid(self,oid):
        self._oid = oid
        
    @property
    def taskguid(self):
        return self._taskguid
    @taskguid.setter
    def taskguid(self,taskguid):
        self._taskguid = taskguid

    @property
    def documentid(self):
        return self._documentid
    @documentid.setter
    def documentid(self,documentid):
        self._documentid = documentid
        
    @property
    def profile(self):
        return self._profile
    @profile.setter
    def profile(self,profile):
        self._profile = profile

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in six.iteritems(self.fieldtypes):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

class DB_Tools(object):
    """
    Class to manage connections to databases and typical processes
    """
    def __init__(self,dbserver,dbname):
        config = Configuration()          
        self.logger = config.packagelogger
        self.logger.debug("initiated db module")
        self.connstr = "Driver={SQL Server};" + """Server={0};Database={1};
            Trusted_Connection=True;
            APP=Assetic Asset Sync""".format(dbserver,dbname)
            #;Charset=utf-8

    @contextmanager
    def open_db_connection(self, commit=False):
        """
        Use contextmanager to manage connections so we cleanup connection
        as we go to avoid database locking issues on exception
        """
        try:
            self.logger.debug("in open db connection")
            connection = pyodbc.connect(self.connstr)
            if sys.version_info < (3,0):
                connection.setencoding(str, 'utf-8')
                connection.setencoding(unicode, 'utf-16le')
                connection.setdecoding(pyodbc.SQL_CHAR, encoding='utf-8')
                connection.setdecoding(pyodbc.SQL_WCHAR, encoding='utf-16le')
                #connection.setencoding(str, encoding='utf-8')
                #connection.setencoding(unicode, encoding='utf-8', ctype=pyodbc.SQL_WCHAR)

            cursor = connection.cursor()
            yield cursor
        except pyodbc.DatabaseError as err:
            cursor.rollback()
            self.logger.error(str(err),exc_info=True)
            raise err
        except UnicodeDecodeError as ex:
            self.logger.error(str(err),exc_info=True)
            cursor.rollback()
            raise ex            
        except Exception as ex:
            cursor.rollback()
            self.logger.error(ex,exc_info=True)
            raise ex
        else:
            if commit:
                cursor.commit()
            else:
                cursor.rollback()
        finally:
            connection.close()
                        
    def data_sync_to_db(self,allrows,columns,dbtable,keyfield,logtable):
        """
        For a array containing csv delimted data sync to db.
        Assums data in columns parameter is the headers for the csv array
        Param1 allrows: the array of data to sync
        Param2 columns: the list of column headers in same order as data
        Param3 dbtable: the database table to sync with
        Param4 keyfield: the primary key used to determine if new or update
        Param5 logtable: the name of the table to log inserts,updates,deletes
        Returned: 0= Success (no db error) else 1 (Exception)
        """
    
        #name for tmp table we'll write data dump to
        tmpdbtable = "#{0}_tmp".format(dbtable)

        #get list of columns in target table (if it exists)
        returncode,existingcolumns = self.get_columns_for_table(dbtable)
        if returncode != 0:
            return returncode
        if len(existingcolumns) == 0:
            ##no existing output table so match to columns passed in
            #use [:] (strip) to instead of columns.copy() for 2.7 compatibility
            existingcolumns = columns[:]

        #get list of columns in log table (if it exists)
        returncode,existinglogcolumns = self.get_columns_for_table(logtable)
        if returncode != 0:
            return returncode
        
        if len(existinglogcolumns) == 0:
            #no existing log table so set to be same as target table
            #use [:] (strip) to instead of .copy() for 2.7 compatibility            
            existinglogcolumns = existingcolumns[:]

        #make sure key field in target table
        #unicode lower() handled differently in PY2 and PY3
        keyexists = True
        if six.PY3 == True:
            if (keyfield.lower().strip('[').strip(']') in map(
            str.lower, existingcolumns)) == False:
                keyexists = False
        else:
            if (keyfield.lower().strip('[').strip(']') in map(
            string.lower, existingcolumns)) == False:
                keyexists = False
                
        if keyexists == False:        
            msg=("Key field '{0}' is not a column in the table '{1}', task will"
                " not be processed").format(keyfield,dbtable)
            self.logger.error(msg)
            return 1
        
        #make sure key field in log table
        #unicode lower() handled differently in PY2 and PY3
        keyexists = True
        if six.PY3 == True:
            if (keyfield.lower().strip('[').strip(']') in map(
            str.lower, existinglogcolumns)) == False:
                keyexists = False
        else:
            if (keyfield.lower().strip('[').strip(']') in map(
            string.lower, existinglogcolumns)) == False:
                keyexists = False
                
        if keyexists == False:
            msg=("Key field [{0}] is not a column in the table [{1}], task will"
            " not be processed").format(keyfield,logtable)
            self.logger.error(msg)
            return 1
        else:
            #don't need key field for log data output so remove from list
            for chk in existinglogcolumns:
                if chk.lower() == keyfield.lower().strip('[').strip(']'):
                    existinglogcolumns.remove(chk)
                    break

        ##get the database table column names and build the various tables
        insertcolumns = []
        diffinsertcolumns = []
        diffupdatecolumns = []
        origvaluecols = []
        createtmpcolsbyselect = []
        logcreatecolumns = []
        logoutputcolumns = []
        #add a "?" for each column to use in sql parameter settings
        parammark = []

        #'columns' is the complete list of columns in the source data
        for column in columns:
            insertcolumns.append("[{0}]".format(column))
            parammark.append("?")
            if column in existingcolumns:
                diffinsertcolumns.append("[{0}]".format(column))
                diffupdatecolumns.append("a.[{0}]=c.[{0}]".format(column))
                createtmpcolsbyselect.append("[{0}]".format(column))
                if column.lower() != keyfield.lower().strip('[').strip(']'):
                    logcreatecolumns.append("[{0}]".format(column))
            else:
                ##this column will be loaded in tmp table, but not synced
                msg = ("column [{0}] in export data is not in table [{1}] "
                       "and will be excluded").format(column,dbtable)
                self.logger.warning(msg)
                ##set a default datatype since it's not in target table
                createtmpcolsbyselect.append("convert(nvarchar(500),'') as [{0}]"
                        .format(column))
                
        for logcolumn in existinglogcolumns:
            if (logcolumn in columns and logcolumn in existingcolumns):
                logoutputcolumns.append("[{0}]".format(logcolumn))
            else:
                if (logcolumn not in ['action','syncdate']):
                    ##this column in log table will be ignored
                    msg = ("column [{0}] in log table is not in either the "
                       "export data or the target table [{1}] and so it "
                       "will be ignored").format(
                           logcolumn,dbtable)
                    self.logger.warning(msg)
                
        sqltblchktmp =  "select object_id('{0}')".format(tmpdbtable)
        sqltruncatetmp = "drop table {0}".format(tmpdbtable)
        sqlcreatetmp = """select {0} into {1} from {2} where 1=0""".format(
            ",".join(createtmpcolsbyselect),tmpdbtable,dbtable)
        
        sqltblchk =  "select object_id('{0}')".format(dbtable)
        sqlcreate = """create table {0} ({1} nvarchar(500))
            """.format(dbtable,"nvarchar(500),".join(insertcolumns ))
        
        sqllogchk = "select object_id('{0}')".format(logtable)
        sqllogcreate = """select convert([nvarchar](10),'') as action,
            {1},convert([datetime],getdate()) as [syncdate],{2} into {0}
            from {3} where 1=0""".format(
                logtable,keyfield,
                ",".join(logcreatecolumns),dbtable)

        ##Now create insert string
        sqlinsert = "insert into {0} (".format(tmpdbtable)
        sqlinsert = sqlinsert + ",".join(insertcolumns ) + ") values ("
        sqlinsert = sqlinsert + ",".join(parammark ) + ")"

        ##Now add data to sql insert string and do insert
        ##(using 'insertmany' effectively row by row)                   
        ##now refresh base table from temp table where there are changes
        ##write the change to the archive file
        ##TODO output all columns for updated record (original record)
        sqldiff = """
        WITH diff as ( select {2}
            FROM {0}
            except select {2}
            FROM {1}
            )
            MERGE {1} as a using diff as c on
            (c.{5}=a.{5})
            when MATCHED then update set {3}
            When not matched then insert
            ({2}) 
            values ({2})
            output $action,inserted.{5},getdate(),deleted.{6} into {4}
            ([action],{5},[syncdate],{7});
        """.format(tmpdbtable,dbtable,",".join(diffinsertcolumns),
                   ",".join(diffupdatecolumns),logtable,keyfield,
                   ",deleted.".join(logoutputcolumns),
                   ",".join(logoutputcolumns))

        ##Now execute all the sql statements in the one transaction block
        ##Bit slow opening & closing db connection, but transaction block
        ##works well
        try:
            with self.open_db_connection(True) as cursor:
                #check if permanent table exists
                cursor.execute(sqltblchk)
                self.logger.debug(sqltblchk)
                row = cursor.fetchone()
                if row[0] == None:
                    #table not found - create permament table for current data
                    self.logger.debug(sqlcreate)
                    cursor.execute(sqlcreate)

                #check if temporary table exists
                self.logger.debug(sqltblchktmp)
                cursor.execute(sqltblchktmp)
                row = cursor.fetchone()
                if row[0] == None:
                    self.logger.debug(sqlcreatetmp)
                    cursor.execute(sqlcreatetmp)
                else:
                    self.logger.debug(sqltruncatetmp)
                    cursor.execute(sqltruncatetmp)
                #check if log table exists
                self.logger.debug(sqllogchk)
                cursor.execute(sqllogchk)
                row = cursor.fetchone()
                if row[0] == None:
                    #table not found - create permament table for logging changes
                    self.logger.debug(sqllogcreate)
                    cursor.execute(sqllogcreate)
                ##write all the data to the temp table in db.
                self.logger.debug(sqlinsert)
                cursor.executemany(sqlinsert,allrows)
                ##execute diff statement to populate permanent tables from temp table
                self.logger.debug(sqldiff)
                cursor.execute(sqldiff)
        except Exception as ex:
            self.logger.exception(ex)
            return 1
        return 0

    def get_columns_for_table(self,dbtable):
        """
        Ge list of columns in database table
        param dbtable: The database table
        Returned: Tuple - 0= success (no db error), else 1 and
        results (an array of columns)
        """
        sql = """select column_name from Information_schema.columns 
            where table_name='{0}' order by ORDINAL_POSITION asc""".format(
                dbtable)
        returncode,results = self.execute_select(sql)
        columnarr  = []
        for result in results:
            ##results come back as a tuple even though only 1 column per record
            columnarr.append(result[0])
        return returncode,columnarr
    
    def execute_select(self,sql):
        """
        Helper method to run an SQL select
        param sql: SQL select statement
        Returned: Tuple - 0= success (no db error), else 1 and
        results (an array of results)
        """
        self.logger.debug(sql)
        try:
            with self.open_db_connection(True) as cursor:
                cursor.execute(sql)
                results = cursor.fetchall()
        except Exception as ex:
            self.logger.exception(ex)
            return 1, None

        return 0, results                

    def execute_single(self,sql,datarow=None):
        """
        Helper method to run a single record SQL insert, update, or delete
        param sql: the sql statement
        param datarow is optional parameter.  Has data as tuple for sql
        Returned: 0= Success (no db error) else 1 (Exception)
        """
        self.logger.debug(sql)
        self.logger.debug(datarow)
        try:
            with self.open_db_connection(True) as cursor:
                if datarow != None:
                    cursor.execute(sql,datarow)
                else:
                    cursor.execute(sql)
        except Exception as ex:
            self.logger.exception(ex)
            return 1
        return 0

    def execute_many(self,sql,datarows):
        """
        Helper method to run a multiple record SQL insert, update, or delete
        Returned: 0= Success (no db error) else 1 (Exception)
        """
        try:
            with self.open_db_connection(True) as cursor:
                cursor.executemany(sql,datarows)
        except Exception as ex:
            self.logger.exception(ex)
            return 1
        return 0

##class UTF8Recoder:
##    def __init__(self, f, encoding):
##        self.reader = codecs.getreader(encoding)(f)
##    def __iter__(self):
##        return self
##    def next(self):
##        return self.reader.next().encode("utf-8")
##
##class UnicodeReader:
##    def __init__(self, f, dialect=csv.excel, encoding="utf-8-sig", **kwds):
##        f = UTF8Recoder(f, encoding)
##        self.reader = csv.reader(f, dialect=dialect, **kwds)
##    def next(self):
##        '''next() -> unicode
##        This function reads and returns the next line as a Unicode string.
##        '''
##        row = self.reader.next()
##        return [unicode(s, "utf-8") for s in row]
##    def __iter__(self):
##        return self
