"""Copyright 2016 Assetic."""

from __future__ import absolute_import
import os
import time
import sys
import base64
import openpyxl  # to read excel files
try:
    import configparser as configparser
except:
    # V2 python?
    import ConfigParser as configparser

# import assetic libraries we'll use
from assetic.api_client import ApiClient
from assetic.configuration import Configuration
from assetic.apis import DataExchangeJobApi
from assetic.apis import DataExchangeTaskApi
from assetic.apis import DocumentApi
from assetic.models import Assetic3IntegrationRepresentationsFilePropertiesRepresentation
from assetic.models import Assetic3IntegrationRepresentationsDocumentRepresentation
from assetic.models import Assetic3IntegrationRepresentationsDataExchangeJobRepresentation
from assetic.rest import ApiException


class BulkProcesses(object):
    """Class to manage bulk processes."""

    def __init__(self, configfile=None):
        """Initialize class with configuration file."""
        if configfile is None:
            configfile = os.path.join(os.getcwd(), 'assetic.ini')

        if configfile is not None and not configfile.isspace() and not os.path.isfile(configfile):
            message = "Configuration file not found.  Expecting assetic.ini in current folder, or set via module init"
            print(message)
            raise ValueError(message)
        else:
            self.configfile = configfile

    def upload_document_photo(self, excelDataFileToRead):
        """Automate document/photo upload in bulk fashion.

        Usage:
            Invoke the script: python.exe -c "import assetic; uploadapi = assetic.BulkProcesses('C:/Users/xyz/assetic.ini'); uploadapi.upload_document_photo('C:/Users/xyz/data.xlsx')"
        Args:
            excelDataFileToRead: XLSX/XLSM/XLTX/XLTM file containing document/photos data to import;
                                 Columns required are as follows (in EXACT ORDER)
                                 Asset Id | Document/Photos Complete File Path | Is Key-photo (TRUE/FALSE)
        Returns: Uploaded document/photos status is written back to excel file specified in excelDataFileToRead
        """
        # Read ini
        settings = configparser.ConfigParser()
        configfile = self.configfile
        settings.read(configfile)
        # Apply config settings
        conf = Configuration()
        conf.host = settings.get('environment', 'url')
        conf.username = settings.get('auth', 'username')
        conf.password = settings.get('auth', 'api_key')
        auth = conf.get_basic_auth_token()

        client = ApiClient(conf.host, "Authorization", auth)
        docapi = DocumentApi(client)

        print('File for bulk document upload operation is specified: {0}\n'.format(
            excelDataFileToRead))
        validDocumentToImportExtensions = ["xlsx", "xlsm", "xltx", "xltm"]
        validImageExtensions = ["jpg", "png", "gif"]

        if excelDataFileToRead is not None and not excelDataFileToRead.isspace() and os.path.exists(excelDataFileToRead) and excelDataFileToRead.lower().split('.')[1] in validDocumentToImportExtensions:
            try:
                wb = openpyxl.load_workbook(excelDataFileToRead)
                sheet = wb.active
                sheet['D' + str(1)] = 'Migrated Document Id'
                sheet['E' + str(1)] = 'Operation Performed On'
                sheet['F' + str(1)] = 'Status'
                wb.save(excelDataFileToRead)
            except Exception as e:
                print('{0}\n'.format(e))
            else:
                for row in range(2, sheet.max_row + 1):
                    assetId = sheet['A' + str(row)].value
                    fileNameToUpload = sheet['B' + str(row)].value
                    isKeyPhotoColumnValue = sheet['C' + str(row)].value
                    isKeyPhoto = False
                    if isKeyPhotoColumnValue:
                        isKeyPhoto = isKeyPhotoColumnValue
                    print('Processing {2}/{3}: Complex Asset Id: [{0}]; File name: [{1}]'.format(
                        assetId, fileNameToUpload, row, sheet.max_row))
                    if fileNameToUpload is not None and not fileNameToUpload.isspace() and os.path.exists(fileNameToUpload):
                        if fileNameToUpload.lower().split('.')[1] in validImageExtensions:
                            if isKeyPhoto:
                                isKeyPhoto = True
                            else:
                                isKeyPhoto = False
                            documentGroupId = 7
                        else:
                            isKeyPhoto = False
                            documentGroupId = 10
                        fileToUploadPath = os.path.abspath(fileNameToUpload)
                        fileToUploadName = os.path.basename(fileNameToUpload)
                        fileExtention = os.path.splitext(
                            fileToUploadPath)[1][1:]
                        with open(fileToUploadPath, "rb") as f:
                            filename = f.name
                            data = f.read()

                            if sys.version_info < (3, 0):
                                filecontents = data.encode("base64")
                            else:
                                filecontents = base64.b64encode(data)
                                filecontents = filecontents.decode(
                                    encoding="utf-8", errors="strict")

                            filesize = os.path.getsize(filename)

                            file_properties = Assetic3IntegrationRepresentationsFilePropertiesRepresentation()
                            file_properties.name = fileToUploadName
                            file_properties.file_size = filesize
                            file_properties.mimetype = fileExtention
                            file_properties.filecontent = filecontents
                            filearray = [file_properties]

                            try:
                                document = Assetic3IntegrationRepresentationsDocumentRepresentation()
                                document.is_key_photo = isKeyPhoto
                                document.document_group = documentGroupId
                                document.parent_type = 'ComplexAsset'
                                document.parent_identifier = assetId
                                document.label = 'Importing complex asset document/photo - [' + fileToUploadName + '] on ' + time.strftime("%d/%m/%Y %H:%M:%S")
                                document.file_property = filearray
                                document.document_size = filesize
                                document.mime_type = fileExtention

                                doc = docapi.document_post(document)
                            except ApiException as e:
                                print('Error occurred while uploading document/photo: [{0}]; Error status {1}, Reason: {2}\n'.format(fileToUploadName, e.status, e.reason))
                            else:
                                docid = doc[0].get('Id')
                                print('UPLOADED document/photo: [{1}]; Response: [{0}]\n'.format(docid, fileToUploadName))

                                isDocumentUploadSuccessful = True if docid is not None and not docid.isspace() else False
                                if isDocumentUploadSuccessful:
                                    sheet['D' + str(row)] = docid
                                    sheet['E' + str(row)] = time.strftime("%c")
                                    sheet['F' + str(row)] = 'SUCCESS'
                                else:
                                    sheet['D' + str(row)] = 'N/A'
                                    sheet['E' + str(row)] = time.strftime("%c")
                                    sheet['F' + str(row)] = 'Error; Please check integration application logs for details'
                    else:
                        print('[{1}/{2}]: File: [{0}] is non-existing\n'.format(fileNameToUpload, row, sheet.max_row))
                        sheet['D' + str(row)] = 'N/A'
                        sheet['E' + str(row)] = time.strftime("%c")
                        sheet['F' + str(row)] = 'Document/Photo file is non-existing'

                    wb.save(excelDataFileToRead)
        else:
            print('File: [{0}] is either not specified or not of correct format (Supported formats are: .xlsx,.xlsm,.xltx,.xltm)\n'.format(
                excelDataFileToRead))

    def dataexchange_upload(self, excelDataFileToRead):
        """Request data-exchange(Bulk data import) requests to SaaS application.

        Usage:
            Invoke the script: python.exe -c "import assetic; uploadapi = assetic.BulkProcesses('C:/Users/xyz/assetic.ini'); uploadapi.dataexchange_upload('C:/Users/xyz/data.xlsx')"
        Args:
            excelDataFileToRead: XLSX/XLSM/XLTX/XLTM file containing data-exchange profile and data to import;
                                 Columns required are as follows (in EXACT ORDER)
                                 Data-exchange Profile Id | Document to Import (Complete file-path)
        Returns: Requested data-exchange task status is written back to excel file specified in excelDataFileToRead
        """
        # Read ini
        settings = configparser.ConfigParser()
        configfile = self.configfile
        settings.read(configfile)
        # Apply config settings
        conf = Configuration()
        conf.host = settings.get('environment', 'url')
        conf.username = settings.get('auth', 'username')
        conf.password = settings.get('auth', 'api_key')
        auth = conf.get_basic_auth_token()

        client = ApiClient(conf.host, "Authorization", auth)
        docapi = DocumentApi(client)
        dataexchangeapi = DataExchangeJobApi(client)
        validExcelFileExtensions = ["xlsx", "xlsm", "xltx",  "xltm"]
        validDocumentToImportExtensions = ["csv"]
        validImageExtensions = ["jpg", "png", "gif"]
        if excelDataFileToRead is not None and not excelDataFileToRead.isspace() and os.path.exists(excelDataFileToRead) and excelDataFileToRead.lower().split('.')[1] in validExcelFileExtensions:
            try:
                wb = openpyxl.load_workbook(excelDataFileToRead)
                sheet = wb.active
                sheet['C' + str(1)] = 'Imported Document Id'
                sheet['D' + str(1)] = 'Data-exchange Job Id'
                sheet['E' + str(1)] = 'Operation Performed On'
                sheet['F' + str(1)] = 'Status'
                wb.save(excelDataFileToRead)
            except Exception as e:
                print('{0}\n'.format(e))
            else:
                totalItems = sheet.max_row - 1
                for row in range(2, sheet.max_row + 1):
                    itemCounter = row - 1
                    dataExchangeProfileId = sheet['A' + str(row)].value
                    documentFilePathToImport = sheet['B' + str(row)].value

                    if documentFilePathToImport is not None and not documentFilePathToImport.isspace() and documentFilePathToImport.lower().split('.')[1] in validDocumentToImportExtensions and os.path.exists(documentFilePathToImport):
                        with open(documentFilePathToImport, "rb") as f:
                            data = f.read()

                        if sys.version_info < (3, 0):
                            filecontents = data.encode("base64")
                        else:
                            filecontents = base64.b64encode(data)
                            filecontents = filecontents.decode(
                                encoding="utf-8", errors="strict")

                        filename, fileExtention = os.path.splitext(
                            documentFilePathToImport)  # [1][1:]
                        fileExtention = fileExtention[1:]
                        filesize = os.path.getsize(documentFilePathToImport)

                        file_properties = Assetic3IntegrationRepresentationsFilePropertiesRepresentation()
                        file_properties.name = filename
                        file_properties.file_size = filesize
                        file_properties.mimetype = fileExtention
                        file_properties.filecontent = filecontents
                        filearray = [file_properties]

                        try:
                            document = Assetic3IntegrationRepresentationsDocumentRepresentation()
                            document.label = 'Importing data-exchange document - [' + filename + '] on ' + time.strftime("%d/%m/%Y %H:%M:%S")
                            document.file_property = filearray
                            document.document_size = filesize
                            document.mime_type = fileExtention

                            doc = docapi.document_post(document)
                        except ApiException as e:
                            print('[{3}/{4}]: Error occurred while uploading data-exchange document: [{0}]; Error status {1}, Reason: {2}\n'.format(filename, e.status, e.reason, itemCounter, totalItems))
                        else:
                            docid = doc[0].get('Id')
                            print('[{3}/{4}]: Processed document: [{0}] upload request for data-exchange profile: {1}; Response: [{2}]\n'.format(filename, dataExchangeProfileId, docid, itemCounter, totalItems))

                            isDocumentUploadSuccessful = True if docid is not None and not docid.isspace() else False
                            if isDocumentUploadSuccessful:
                                try:
                                    job = Assetic3WebApiModelsDataExchangeDataExchangeJobVM()
                                    job.profile_id = dataExchangeProfileId
                                    job.document_id = docid
                                    task = dataexchangeapi.data_exchange_job_post(
                                        job)
                                except ApiException as e:
                                    sheet['C' + str(row)] = docid
                                    sheet['D' + str(row)] = 'N/A'
                                    sheet['E' + str(row)] = time.strftime("%c")
                                    sheet['F' + str(row)] = 'Error; ' + \
                                        str(e.status) + ': ' + e.reason
                                    print('[{3}/{4}]: Error occurred while requesting data-exchange task: [{0}]; Error status {1}, Reason: {2}\n'.format(dataExchangeProfileId, e.status, e.reason, itemCounter, totalItems))
                                else:
                                    print('[{3}/{4}]: Requested Data-exchange task: {0} for profile: {1}/Document: {2}\n'.format(task, dataExchangeProfileId, filename, itemCounter, totalItems))

                                    sheet['C' + str(row)] = docid
                                    sheet['D' + str(row)] = task
                                    sheet['E' + str(row)] = time.strftime("%c")
                                    sheet['F' + str(row)] = 'SUCCESS'
                            else:
                                sheet['C' + str(row)] = 'N/A'
                                sheet['D' + str(row)] = 'N/A'
                                sheet['E' + str(row)] = time.strftime("%c")
                                sheet['F' + str(row)] = 'Error; Please check integration application logs for details'
                    else:
                        print('[{2}/{3}]: Document to import: [{1}] for data-exchange profile: [{0}] is either non-existing or of not CSV format\n'.format(
                            documentFilePathToImport, dataExchangeProfileId, itemCounter, totalItems))
                        sheet['C' + str(row)] = 'N/A'
                        sheet['D' + str(row)] = 'N/A'
                        sheet['E' + str(row)] = time.strftime("%c")
                        sheet['F' + str(row)] = 'Document to import file is either non-existing or not of CSV format'

                    wb.save(excelDataFileToRead)
        else:
            print('File: [{0}] is either not specified or not of correct format (Supported formats are: .xlsx,.xlsm,.xltx,.xltm)\n'.format(
                excelDataFileToRead))

    def dataexchange_get_error_summary(self, excelDataFileToRead, errorFilesStorageDirectory):
        """Check status of data-exchange task(s) requested in [dataexchange_upload] function.

        Usage:
            Invoke the script: python.exe -c "import assetic; uploadapi = assetic.BulkProcesses('C:/Users/xyz/assetic.ini'); uploadapi.dataexchange_get_error_summary('C:/Users/xyz/data.xlsx', 'C:/Users/xyz/error_files/')"
        Args:
            excelDataFileToRead: XLSX/XLSM/XLTX/XLTM file containing data-exchange profile, document and task id;
                                 Columns required are as follows (in EXACT ORDER)
                                 Data-exchange ProfileId | Document to Import (Complete file-path) | Imported Document Id | Data-exchange Job Id | Operation Performed On | Status
            errorFilesStorageDirectory: Directory location where imported documents are stored

        Returns: Status of data-exchange tasks is written back to excel file specified in excelDataFileToRead and relvant imported document is saved in errorFilesStorageDirectory directory
        """
        # Read ini
        settings = configparser.ConfigParser()
        configfile = self.configfile
        settings.read(configfile)
        # Apply config settings
        conf = Configuration()
        conf.host = settings.get('environment', 'url')
        conf.username = settings.get('auth', 'username')
        conf.password = settings.get('auth', 'api_key')
        auth = conf.get_basic_auth_token()

        client = ApiClient(conf.host, "Authorization", auth)
        docapi = DocumentApi(client)
        dataExchangeTaskapi = DataExchangeTaskApi(client)

        validDocumentToImportExtensions = ["xlsx", "xlsm", "xltx",  "xltm"]
        validImageExtensions = ["jpg", "png", "gif"]
        if excelDataFileToRead is not None and not excelDataFileToRead.isspace() and os.path.exists(excelDataFileToRead) and excelDataFileToRead.lower().split('.')[1] in validDocumentToImportExtensions:
            try:
                wb = openpyxl.load_workbook(excelDataFileToRead)
                sheet = wb.active
                sheet['G' + str(1)] = 'Error Document Id'
                sheet['H' + str(1)] = 'Summary'
                wb.save(excelDataFileToRead)
            except Exception as e:
                print('{0}\n'.format(e))
            else:
                is_all_complete = True
                totalItems = sheet.max_row - 1
                for row in range(2, sheet.max_row + 1):
                    itemCounter = row - 1
                    importedDocumentId = sheet['C' + str(row)].value
                    dataExchangeTaskId = sheet['D' + str(row)].value

                    if dataExchangeTaskId is not None and not dataExchangeTaskId.isspace() and dataExchangeTaskId != 'N/A':
                        try:
                            task = dataExchangeTaskapi.data_exchange_task_get(dataExchangeTaskId)
                        except ApiException as e:
                            sheet['G' + str(row)] = e.status
                            sheet['H' + str(row)] = e.reason
                            print('[{3}]/[{4}]: Error occurred while fetching status of data-exchange task:[{0}]; Error status {1}, Reason: {2}\n'.format(dataExchangeTaskId, e.status, e.reason, itemCounter, totalItems))
                        else:
                            status = task.get('Status')
                            statusDescription = task.get('StatusDescription')
                            print('[{3}]/[{4}]: [{0}]: Data Exchange Job has status {1} [{2}]'.format(dataExchangeTaskId, status, statusDescription, itemCounter, totalItems))
                            # Complete with Error
                            if status == int('4'):
                                errorDocumentId = task.get('ErrorDocumentId')
                                if errorDocumentId is not None and not errorDocumentId.isspace():
                                    print('[{3}/{4}]: Fetching error document: [{0}] for data-exchange task: [{2}] into [{1}]'.format(errorDocumentId, errorFilesStorageDirectory, dataExchangeTaskId, itemCounter, totalItems))
                                    self.document_get(errorFilesStorageDirectory, errorDocumentId)
                                    sheet['G' + str(row)] = errorDocumentId

                                try:
                                    sheet['H' + str(row)] = '[COMPLETED WITH ERROR]: ' + task.get('Summary')
                                except:
                                    sheet['H' + str(row)] = '[COMPLETED WITH ERROR]'

                            # Complete with No Error
                            elif status == int('3'):
                                sheet['G' + str(row)] = 'N/A'
                                try:
                                    sheet['H' + str(row)] = '[COMPLETED WITH NO ERROR]: ' + task.get('Summary')
                                except:
                                    sheet['H' + str(row)] = '[COMPLETED WITH NO ERROR]'

                            else:
                                is_all_complete = False
                                print('[{3}/{4}]: Data-exchange task: [{0}] is in status: [{1}/{2}]\n'.format(dataExchangeTaskId, status, statusDescription, itemCounter, totalItems))
                                sheet['G' + str(row)] = ''
                                try:
                                    sheet['H' + str(row)] = '[{0}]: {1}'.format(statusDescription, task.get('Summary'))
                                except:
                                    print('[{0}/{1}]: Error getting Status data. Please try again later - the upload job is likely still running.', itemCounter, totalItems)
                    else:
                        sheet['G' + str(row)] = 'N/A'
                        sheet['H' + str(row)] = 'Data-exchange task id is null, empty or not valid'

                    wb.save(excelDataFileToRead)

                return is_all_complete
        else:
            print(
                'File: [{0}] is either not specified or not of correct format(Supported formats are: .xlsx,.xlsm,.xltx,.xltm)\n'.format(excelDataFileToRead))

    def document_get(self, fileStorageDirectoryLocation, docid):
        """Download specified document into directory.

        Usage:
            Invoke the script: python.exe -c "import assetic; uploadapi = assetic.BulkProcesses('C:/Users/xyz/assetic.ini'); uploadapi.document_get('C:/Users/xyz/', '1234-34566-23456-2334324')"
        Args:
            fileStorageDirectoryLocation: Directory location to save specified document
            docid: Document Id of document to be downloaded
        """
        # Read ini
        settings = configparser.ConfigParser()
        configfile = self.configfile
        settings.read(configfile)
        # Apply config settings
        conf = Configuration()
        conf.host = settings.get('environment', 'url')
        conf.username = settings.get('auth', 'username')
        conf.password = settings.get('auth', 'api_key')
        auth = conf.get_basic_auth_token()

        client = ApiClient(conf.host, "Authorization", auth, None, 'Content-Disposition')
        docapi = DocumentApi(client)
        if fileStorageDirectoryLocation is not None and not fileStorageDirectoryLocation.isspace() and os.path.exists(fileStorageDirectoryLocation) and docid is not None and not docid.isspace():
            # Stop downloading a file if there's already one with the Document
            # ID stored
            for file in os.listdir(fileStorageDirectoryLocation):
                if docid in file:
                    print('A document with Document ID [{0}] already exists. Ignoring file download.'.format(docid))
                    return False

            try:
                getfile = docapi.document_get_document_file(docid)
            except ApiException as e:
                print('Error occurred while fetching Document:[{0}]; Error status {1}, Reason: {2}\n'.format(docid, e.status, e.reason))
            else:
                getfile = docapi.document_get_document_file(docid)
                if getfile is not None:
                    if 'attachment' in getfile[1] and 'filename=' in getfile[1]:
                        filename = getfile[1].split('filename=', 1)[1]
                        if '"' in filename or "'" in filename:
                            filename = filename[1:-1]
                        fullfilename = os.path.join(
                            fileStorageDirectoryLocation, filename)
                    else:
                        fullfilename = os.path.join(
                            fileStorageDirectoryLocation, docid)

                    data = getfile[0]
                    if sys.version_info >= (3, 0):
                        try:
                            data = data.decode('utf-8')
                        except:
                            data = data  # pass

                    # if file with name already exists then add document id at
                    # the end to make it unique
                    if fullfilename is not None and not fullfilename.isspace():
                        fileNameToModify, fileExtention = os.path.splitext(
                            fullfilename)
                        fullfilename = fileNameToModify + '_' + \
                            docid + '.' + fileExtention[1:]

                        # Need to handle binary and text files (e.g. CSV error
                        # files if Data Exchange jobs fail)
                        try:
                            with open(fullfilename, 'wb') as out_file:
                                out_file.write(data)
                                print('Created file: [{0}]\n'.format(fullfilename))
                        except:
                            try:
                                with open(fullfilename, 'w') as out_file:
                                    out_file.write(data)
                                    print('Created file: [{0}]\n'.format(fullfilename))
                            except:
                                print('Failed getting file [{0}] in binary and plain text modes!'.format(fullfilename))
                else:
                    print('No document fetch response is received')
        else:
            print('Either File storage directory location or document id is not specified or is in incorrect format')
